@@ -101,9 +101,11 @@ Path absPath(PathView path, std::optional<PathView> dir, bool resolveSymlinks)
     return canonPath(path, resolveSymlinks);
 }
 
-std::filesystem::path absPath(const std::filesystem::path & path, bool resolveSymlinks)
+std::filesystem::path
+absPath(const std::filesystem::path & path, const std::filesystem::path * dir_, bool resolveSymlinks)
 {
-    return absPath(path.string(), std::nullopt, resolveSymlinks);
+    std::optional<std::string> dir = dir_ ? std::optional<std::string>{dir_->string()} : std::nullopt;
+    return absPath(PathView{path.string()}, dir.transform([](auto & p) { return PathView(p); }), resolveSymlinks);
 }
 
 Path canonPath(PathView path, bool resolveSymlinks)
@@ -242,10 +244,15 @@ bool pathAccessible(const std::filesystem::path & path)
     }
 }
 
-Path readLink(const Path & path)
+std::filesystem::path readLink(const std::filesystem::path & path)
 {
     checkInterrupt();
-    return std::filesystem::read_symlink(path).string();
+    return std::filesystem::read_symlink(path);
+}
+
+Path readLink(const Path & path)
+{
+    return readLink(std::filesystem::path{path}).string();
 }
 
 std::string readFile(const Path & path)
@@ -253,8 +260,7 @@ std::string readFile(const Path & path)
     AutoCloseFD fd = toDescriptor(open(
         path.c_str(),
         O_RDONLY
-// TODO
-#ifndef _WIN32
+#ifdef O_CLOEXEC
             | O_CLOEXEC
 #endif
         ));
@@ -287,8 +293,7 @@ void readFile(const Path & path, Sink & sink, bool memory_map)
     AutoCloseFD fd = toDescriptor(open(
         path.c_str(),
         O_RDONLY
-// TODO
-#ifndef _WIN32
+#ifdef O_CLOEXEC
             | O_CLOEXEC
 #endif
         ));
@@ -302,8 +307,7 @@ void writeFile(const Path & path, std::string_view s, mode_t mode, FsSync sync)
     AutoCloseFD fd = toDescriptor(open(
         path.c_str(),
         O_WRONLY | O_TRUNC | O_CREAT
-// TODO
-#ifndef _WIN32
+#ifdef O_CLOEXEC
             | O_CLOEXEC
 #endif
         ,
@@ -337,8 +341,7 @@ void writeFile(const Path & path, Source & source, mode_t mode, FsSync sync)
     AutoCloseFD fd = toDescriptor(open(
         path.c_str(),
         O_WRONLY | O_TRUNC | O_CREAT
-// TODO
-#ifndef _WIN32
+#ifdef O_CLOEXEC
             | O_CLOEXEC
 #endif
         ,
@@ -669,18 +672,18 @@ void AutoUnmount::cancel()
 
 //////////////////////////////////////////////////////////////////////
 
-std::string defaultTempDir()
+std::filesystem::path defaultTempDir()
 {
     return getEnvNonEmpty("TMPDIR").value_or("/tmp");
 }
 
-Path createTempDir(const Path & tmpRoot, const Path & prefix, mode_t mode)
+std::filesystem::path createTempDir(const std::filesystem::path & tmpRoot, const std::string & prefix, mode_t mode)
 {
     while (1) {
         checkInterrupt();
-        Path tmpDir = makeTempPath(tmpRoot, prefix);
+        std::filesystem::path tmpDir = makeTempPath(tmpRoot, prefix);
         if (mkdir(
-                tmpDir.c_str()
+                tmpDir.string().c_str()
 #ifndef _WIN32 // TODO abstract mkdir perms for Windows
                     ,
                 mode
@@ -706,11 +709,31 @@ Path createTempDir(const Path & tmpRoot, const Path & prefix, mode_t mode)
     }
 }
 
+AutoCloseFD createAnonymousTempFile()
+{
+    AutoCloseFD fd;
+#ifdef O_TMPFILE
+    fd = ::open(defaultTempDir().c_str(), O_TMPFILE | O_CLOEXEC | O_RDWR, S_IWUSR | S_IRUSR);
+    if (!fd)
+        throw SysError("creating anonymous temporary file");
+#else
+    auto [fd2, path] = createTempFile("nix-anonymous");
+    if (!fd2)
+        throw SysError("creating temporary file '%s'", path);
+    fd = std::move(fd2);
+#  ifndef _WIN32
+    unlink(requireCString(path)); /* We only care about the file descriptor. */
+#  endif
+#endif
+    return fd;
+}
+
 std::pair<AutoCloseFD, Path> createTempFile(const Path & prefix)
 {
-    Path tmpl(defaultTempDir() + "/" + prefix + ".XXXXXX");
+    Path tmpl(defaultTempDir().string() + "/" + prefix + ".XXXXXX");
     // Strictly speaking, this is UB, but who cares...
     // FIXME: use O_TMPFILE.
+    // FIXME: Windows should use FILE_ATTRIBUTE_TEMPORARY | FILE_FLAG_DELETE_ON_CLOSE
     AutoCloseFD fd = toDescriptor(mkstemp((char *) tmpl.c_str()));
     if (!fd)
         throw SysError("creating temporary file '%s'", tmpl);
@@ -720,21 +743,20 @@ std::pair<AutoCloseFD, Path> createTempFile(const Path & prefix)
     return {std::move(fd), tmpl};
 }
 
-Path makeTempPath(const Path & root, const Path & suffix)
+std::filesystem::path makeTempPath(const std::filesystem::path & root, const std::string & suffix)
 {
     // start the counter at a random value to minimize issues with preexisting temp paths
     static std::atomic<uint32_t> counter(std::random_device{}());
-    auto tmpRoot = canonPath(root.empty() ? defaultTempDir() : root, true);
+    auto tmpRoot = canonPath(root.empty() ? defaultTempDir().string() : root.string(), true);
     return fmt("%1%/%2%-%3%-%4%", tmpRoot, suffix, getpid(), counter.fetch_add(1, std::memory_order_relaxed));
 }
 
 void createSymlink(const Path & target, const Path & link)
 {
-    try {
-        std::filesystem::create_symlink(target, link);
-    } catch (std::filesystem::filesystem_error & e) {
-        throw SysError("creating symlink '%1%' -> '%2%'", link, target);
-    }
+    std::error_code ec;
+    std::filesystem::create_symlink(target, link, ec);
+    if (ec)
+        throw SysError(ec.value(), "creating symlink '%1%' -> '%2%'", link, target);
 }
 
 void replaceSymlink(const std::filesystem::path & target, const std::filesystem::path & link)
