@@ -33,12 +33,10 @@ struct CmdAddDerivation : MixDryRun, StoreCommand
     {
         auto json = nlohmann::json::parse(drainFD(STDIN_FILENO));
 
-        auto drv = Derivation::fromJSON(json);
+        auto drv = Derivation::parseJsonAndValidate(*store, json);
 
         auto drvPath = writeDerivation(*store, drv, NoRepair, /* read only */ dryRun);
 
-        drv.checkInvariants(*store, drvPath);
-
         writeDerivation(*store, drv, NoRepair, dryRun);
 
         logger->cout("%s", store->printStorePath(drvPath));
